/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __surface_SurfaceAccessibleBase_h__
#define __surface_SurfaceAccessibleBase_h__

namespace fe
{

/**************************************************************************//**
    @brief Common Functionality for Accessible Surface

	@ingroup surface
*//***************************************************************************/
class FE_DL_EXPORT SurfaceAccessibleBase:
	public ObjectSafeShared<SurfaceAccessibleBase>,
	virtual public SurfaceAccessibleI
{
	public:
								SurfaceAccessibleBase(void);
virtual							~SurfaceAccessibleBase(void);

								//* As Protectable
virtual	Protectable*			clone(Protectable* pInstance=NULL);

								//* as SurfaceAccessibleI

virtual	void					bind(Instance a_instance)
								{
									Record record=a_instance.cast<Record>();
									bind(record);
									if(record.isValid())
									{
										sp<Scope> spScope=
												record.layout()->scope();
										bind(spScope);
									}
								}
virtual	BWORD					isBound(void)
								{	return m_weakRecord.isValid(); }

virtual	void					bind(sp<Scope> a_spScope)
								{	m_spScope=a_spScope;
									reset(); }
virtual	BWORD					load(String a_filename);
virtual	BWORD					load(String a_filename,Real a_frame,
										String a_options,
										sp<StringFilterI> a_spFilter)
								{	return load(a_filename); }
virtual	BWORD					save(String a_filename);

virtual	void					bind(sp<SurfaceAccessibleI::ThreadingState>
										a_spThreadingState)
								{	m_spThreadingState=a_spThreadingState; }
virtual	void					lock(void)		{ safeLock(); }
virtual	void					unlock(void)	{ safeUnlock(); }

virtual	void					groupNames(Array<String>& a_rNameArray,
										SurfaceAccessibleI::Element
										a_element) const;
virtual	void					attributeSpecs(
										Array<SurfaceAccessibleI::Spec>&
										a_rSpecs,
										SurfaceAccessibleI::Element
										a_element) const;

virtual	void					clear(void)		{ reset(); }

virtual void					require(SurfaceAccessibleI::Element a_element,
										String a_name,String a_typeName);

virtual	I32						count(SurfaceAccessibleI::Element
										a_element) const;

virtual sp<SurfaceAccessorI>	accessor(Element a_element,String a_name,
										Creation a_create);

virtual sp<SurfaceAccessorI>	accessor(Element a_element,
										Attribute a_attribute,
										Creation a_create);

virtual sp<SurfaceAccessorI>	accessor(Element a_element,String a_name)
								{	return accessor(a_element,a_name,
											e_createMissing); }

virtual sp<SurfaceAccessorI>	accessor(Element a_element,
										Attribute a_attribute)
								{	return accessor(a_element,a_attribute,
											e_createMissing); }

virtual	BWORD					discard(SurfaceAccessibleI::Element a_element,
										String a_name)
								{	return FALSE; }
virtual	BWORD					discard(SurfaceAccessibleI::Element a_element,
										SurfaceAccessibleI::Attribute
										a_attribute);

virtual	sp<SurfaceI>			surface(void)
								{	return surface(String()); }
virtual	sp<SurfaceI>			surface(String a_group);
virtual	sp<SurfaceI>			subSurface(U32 a_subIndex)
								{	return subSurface(a_subIndex,String()); }
virtual	sp<SurfaceI>			subSurface(U32 a_subIndex,String a_group);

virtual	void					copy(sp<SurfaceAccessibleI>
										a_spSurfaceAccessible)
								{	reset();
									append(a_spSurfaceAccessible,NULL); }

virtual	void					copy(sp<SurfaceAccessibleI>
										a_spSurfaceAccessible,
										sp<FilterI> a_spFilter)
								{	reset();
									append(a_spSurfaceAccessible,
											NULL,a_spFilter); }

virtual	void					append(sp<SurfaceAccessibleI>
										a_spSurfaceAccessible,
										const SpatialTransform*
										a_pTransform)
								{	append(a_spSurfaceAccessible,
											a_pTransform,sp<FilterI>(NULL)); }

								//* promote?
		void					append(sp<SurfaceAccessibleI>
										a_spSurfaceAccessible,
										const SpatialTransform*
										a_pTransform,
										sp<FilterI> a_spFilter);

virtual	void					instance(sp<SurfaceAccessibleI>
										a_spSurfaceAccessible,
										const Array<SpatialTransform>&
										a_rTransformArray)					{}

								//* default implementation is non-atomic
virtual	sp<SpannedRange>		atomize(AtomicChange a_atomicChange,
										String a_group,U32 a_desiredCount);

virtual	sp<SpannedRange>		atomizeConnectivity(AtomicChange a_atomicChange,
										String a_group,U32 a_desiredCount,
										BWORD a_checkPages);

virtual void					setPaging(BWORD a_paging)
								{	m_paging=a_paging; }
virtual BWORD					paging(void) const	{ return m_paging; }

virtual void					preparePaging(AtomicChange a_atomicChange,
										String a_group)						{}
virtual I32						pointPage(U32 a_pointIndex) const
								{	return 0; }
virtual I32						primitivePage(U32 a_primitiveIndex) const
								{	return 0; }

								//* as SurfaceAccessibleBase
		void					bind(Record& a_rRecord)
								{	m_weakRecord=a_rRecord; }
		Record					record(void)
								{	return m_weakRecord; }

static	String					elementLayout(
										SurfaceAccessibleI::Element a_element);

	protected:
virtual	void					reset(void);

static	String					commonName(String a_attrName);
static	String					attributeName(String a_specName);

		SurfaceAccessibleI::Threading	threading(void) const
							{
								if(m_spThreadingState.isNull())
								{
									return SurfaceAccessibleI::e_unknown;
								}
								return Threading(
										m_spThreadingState->threading());
							}

		sp<SurfaceAccessibleI::ThreadingState>	m_spThreadingState;

		sp<Scope>				m_spScope;
		sp<RecordGroup>			m_spRecordGroup;
		WeakRecord				m_weakRecord;
		BWORD					m_paging;
};

} // namespace

#endif /* __surface_SurfaceAccessibleBase_h__ */
