import string
import re
import os
import sys
import shutil
import copy
from distutils.core import setup as dist_setup, Extension
from distutils.util import get_platform
import distutils.sysconfig
forge = sys.modules["forge"]
sys.path.append(forge.rootPath)
import common
import local

config = {}

###############################################################################
#### CHECK PYTHON VERSION #####################################################
version_match = re.compile(r'(\d*\.\d*).*').match(sys.version)
if not version_match:
	forge.cprint(forge.RED,1,'ERROR: cannot determine python version')
else:
	forge.python_version = float(version_match.group(1))
	if forge.python_version < 2.3:
		forge.cprint(forge.RED,1,'ERROR: python >= 2.3 required')
		forge.failed = 1

###############################################################################
#### GENERAL ##################################################################
config['EXTENSIONS'] = []

if forge.codegen == 'validate':
	config['FE_CODEGEN'] = 'FE_VALIDATE'
	config['_REENTRANT'] = None

if forge.codegen == 'debug':
	config['FE_CODEGEN'] = 'FE_DEBUG'
	config['_REENTRANT'] = None

if forge.codegen == 'profile':
	config['FE_CODEGEN'] = 'FE_PROFILE'

if forge.codegen == 'optimize':
	config['FE_CODEGEN'] = 'FE_OPTIMIZE'

forge.basiclibs = [	"fePlatformLib",
					"feMemoryLib",
					"feCoreLib" ]

if forge.api == "x86_win32" or forge.api == "x86_win64":
	forge.basiclibs += [ "feImportMemoryLib" ]

forge.corelibs = forge.basiclibs + [
					"fePluginLib",
					"feDataLib" ]

forge.libRoot = os.getenv('FE_LIB')
if not forge.libRoot:
	forge.libRoot=forge.rootPath

forge.objRoot = os.getenv('FE_OBJ')
if not forge.objRoot:
	forge.objRoot=forge.libRoot

config['FE_DOUBLE_REAL'] = '1'
if os.environ.has_key("FE_DOUBLE_REAL"):
	config['FE_DOUBLE_REAL'] = os.environ["FE_DOUBLE_REAL"]

config['FE_USE_SSE'] = '0'
if forge.fe_config.has_key("FE_USE_SSE"):
	config['FE_USE_SSE'] = forge.fe_config['FE_USE_SSE']

config['FE_FORGE_CXX'] = '"%s"' % forge.cxx


config['FE_USE_VISIBILITY'] = '0'
if os.environ.has_key("FE_USE_VISIBILITY"):
	config['FE_USE_VISIBILITY'] = os.environ["FE_USE_VISIBILITY"]

config['FE_MT'] = '1'
if os.environ.has_key("FE_MT"):
	config['FE_MT'] = os.environ["FE_MT"]

#pcre_check = os.popen('pcre-config --version','r')
#pcre_version = pcre_check.read()
#pcre_check.close()
#if pcre_version == "":
#	pcre_version = 0
#	forge.cprint(forge.CYAN,0,'PCRE version undetermined')
#else:
#	pcre_version = float(pcre_version)
#	forge.cprint(forge.CYAN,0,'PCRE version ' + str(pcre_version))
#
#if os.environ.has_key("FE_BOOST_REGEX") and int(os.environ["FE_BOOST_REGEX"]):
#	forge.cprint(forge.CYAN,0,'using boost regular expressions')
#	forge.fe_re = 'FE_RE_BOOST'
#elif os.path.exists('/usr/include/pcrecpp.h') and pcre_version >= 0:
#	forge.cprint(forge.CYAN,0,'using PCRE regular expressions')
#	forge.fe_re = 'FE_RE_PCRE'
#else:
#	forge.cprint(forge.CYAN,0,'using no regular expressions')
#	forge.fe_re = 'FE_RE_NONE'

###############################################################################
#### PLATFORMS ################################################################
common.platform_setup()

if forge.api == "x86_linux" or forge.api == "x86_64_linux" or forge.api=="x86_osx" or forge.api == "ppc_osx" or  forge.api == "ppc_linux":
	if forge.cpp_0x:
		config['GNU_UNORDERED_MAP'] = None
		forge.cprint(forge.YELLOW,0,'using unordered_map')
	else:
		config['GNU_STL_HASH_MAP'] = None
#		forge.cprint(forge.CYAN,0,'using hash_map')

elif forge.fe_os == "FE_CYGWIN":
	config['GNU_STL_HASH_MAP'] = None

elif forge.fe_os == "FE_WIN32" or forge.fe_os == "FE_WIN64":
	pass

###############################################################################
#### COMPILERS ################################################################
common.compiler_setup()

config['FE_OS'] = forge.fe_os
config['FE_OSVER'] = forge.fe_osver
config['FE_COMPILER'] = forge.fe_compiler
config['FE_2DGL'] = forge.fe_gl2d
config['FE_3DGL'] = forge.fe_gl3d
config['FE_HW'] = forge.fe_hw
#config['FE_RE'] = forge.fe_re

forge.objPath = os.path.join(forge.objRoot, 'obj', forge.apiCodegen)
forge.libPath = os.path.join(forge.libRoot, 'lib', forge.apiCodegen)
forge.binPath = os.path.join(forge.rootPath, 'bin', forge.api)
forge.srcPath = os.path.join(forge.rootPath, 'src')

#forge.includemap["master_headers"] = "fe"

forge.includemap["root"] = forge.srcPath

###############################################################################
#### BOOST ####################################################################
#common.setup_boost()

############################################################################
# print project build header
#forge.cprint(forge.CYAN,0,'compiler = ' + forge.fe_compiler)
forge.cprint(forge.CYAN,0,'Python ' + str(forge.python_version))
forge.cprint(forge.CYAN,0,'building ' + forge.apiCodegen)

###############################################################################
#### LOCAL TWEAKS #############################################################

local.local_setup()

# creates forge.houdini_sources and adds to forge.boost_sources
common.scan_houdini_installs()

common.scan_maya_installs()

# LD_LIBRARY_PATH
def prepend_to_ld_lib_path(ld_lib_path):
	old_ld_lib_path = os.getenv('LD_LIBRARY_PATH')
	if old_ld_lib_path:
		ld_lib_path += ':' + old_ld_lib_path
	os.environ['LD_LIBRARY_PATH'] = ld_lib_path

prepend_to_ld_lib_path(forge.libPath)

# PYTHONPATH
old_pythonpath = os.getenv('PYTHONPATH')
pythonpath = forge.libPath
if old_pythonpath:
	pythonpath += ':' + old_pythonpath
os.environ['PYTHONPATH'] = pythonpath

# INCLUDEPATH
if os.environ.has_key('INCLUDEPATH'):
	forge.cprint(forge.CYAN,0,'INCLUDEPATH:')
	include_path = string.split(os.environ['INCLUDEPATH'], ':')
	for include in include_path:
		forge.cprint(forge.CYAN,0,'  ' + include)
		forge.includemap['~_' + include] = include
#else:
#	forge.cprint(forge.CYAN,0,'INCLUDEPATH not set')

# BOOST VERSION
#forge.cprint(forge.CYAN,0,'BOOST_VERSION: ' + forge.boost_version)

# set HASH_MAP_IS_MAP to force ext::hash_map to be std::map
if os.getenv('HASH_MAP_IS_MAP'):
	config['HASH_MAP_IS_MAP'] = None
	forge.cprint(forge.CYAN,0,'HASH_MAP_IS_MAP set')
# automatically done for win32
if forge.fe_os == "FE_WIN32" or forge.fe_os == "FE_WIN64":
	config['HASH_MAP_IS_MAP'] = None

forge.cprint(forge.YELLOW,1,'Free Electron')

############################################################################
test_file = """
#include <string.h>
int main(void)
{
	char s[] = "hello";
	memrchr((void *)s, 'e', 5);
	return (0);
}
\n"""

if not forge.cctest(test_file):
	config['HAS_MEMRCHR'] = None


############################################################################
# config.h
#configname=os.path.join(forge.srcPath, 'fe','config.h')
config_dir = os.path.join(forge.libPath, 'fe')
if not os.path.exists(config_dir):
	os.makedirs(config_dir)
configname=os.path.join(config_dir, 'config.h')
forge.includemap['headers_in_lib'] = os.path.join(forge.libPath)
config_header = forge.FileTarget(configname)
config_header.AddDep(forge.targetRegistry['ROOT_BUILD_FILE'])
forge.targetRegistry['root'].AddDep(config_header)
def config_text():
	txt = ""
	txt += "/* Free Electron Configuration File */\n"
	for define in config.keys():
		txt += '\n#ifndef %s\n' % (define)
		if config[define] == None:
			txt += '#define %s\n' % (define)
		else:
			txt += '#define %-30s %s\n' % (define, config[define])
		txt += '#endif\n'
	return txt
	
def generate_config(target):
	forge.cprint(forge.WHITE,0,'generating config file')
	f = file(target.name, "w")
	txt = config_text()
	f.write(txt)
	f.close()

# if it does not exist at all, just build it.  Dependencies do not generate
# off of the header files to not existant headers, so we need to force it here
if not os.path.exists(configname):
	generate_config(config_header)
# if it changed content, build it
else:
	fp = open(configname)
	txt0 = fp.read()
	fp.close()
	txt1 = config_text()
	if txt0 != txt1:
		generate_config(config_header)

#configname="(.*)config.h"
configname = os.path.join(forge.srcPath,'fe','config.h')
forge.defaultContext.AppendRule(configname, generate_config)

############################################################################
# attribute_define.h
#attrtxtname=os.path.join(forge.rootPath, 'doc', 'attribute_names.txt')
#attrtxt = forge.FileTarget(attrtxtname)
#attrdefname=os.path.join(config_dir, 'attribute_define.h')
#attrdef_header = forge.FileTarget(attrdefname)
#attrdef_header.AddDep(attrtxt)
#forge.targetRegistry['root'].AddDep(attrdef_header)
#forge.targetRegistry['root'].AddDep(attrtxt)
#def generate_attrdef(target):
#	forge.cprint(forge.WHITE,0,'generating attribute_define file')
#	attrdef_cmd = os.path.join(forge.rootPath, 'doc', 'attribute_define.py')
#	attrdef_cmd += ' %s %s' % (attrtxtname, attrdefname)
#	os.system(attrdef_cmd)

#if not os.path.exists(attrdefname):
#	generate_attrdef(attrdef_header)

#attrdefnm="(.*)attribute_define.h"
#forge.defaultContext.AppendRule(attrdefnm, generate_attrdef)


############################################################################

def python_build_name():
	v = sys.version_info
	return get_platform() + "-" + str(v[0]) + "." + str(v[1])

# 20130928 jweber deactivated
if False:# forge.fe_os == "FE_LINUX":
	forge.cprint(forge.CYAN,0,'effective LD_LIBRARY_PATH:')
	ld_lib_path = string.split(os.environ['LD_LIBRARY_PATH'], ':')
	forge.linkmap['LD_LIBRARY_PATH'] = ''
	for l in ld_lib_path:
		forge.linkmap['LD_LIBRARY_PATH'] += ' -L' + l
		if os.path.exists(l):
			forge.cprint(forge.CYAN,0,'  ' + l)
		else:
			forge.cprint(forge.RED,0,'  ' + l + '  [does not exist]')

def scrub_tree(arg,dirname,names):
	# removals
	extensions = [ 'pyc' ]
	for ext in extensions:
		for filename in names:
			if re.compile(r'.*\.%s' % (ext)).match(filename):
				fullname = os.path.join(dirname,filename)
				os.remove(fullname)
	# warnings
	extensions = [ 'swp' ]
	for ext in extensions:
		for filename in names:
			if re.compile(r'.*\.%s' % (ext)).match(filename):
				fullname = os.path.join(dirname,filename)
				forge.cprint(forge.YELLOW,0,'warning: '+ fullname)


class ScrubTarget(forge.Target):
	def __init__(self,str):
			forge.Target.__init__(self,str)
			self.help = """\
Scrub transient files in a sledge hammer fashion (the 'clean' target
does so in a more surgical fashion.)

However, this is probably more analagous to 'make clean', so if that is what
you really want then this is your target.
"""

	def Build(self, reference_time):
		directories = [	forge.objPath, forge.libPath ]
		for dir in directories:
			if os.path.exists(dir) and os.path.isdir(dir):
				forge.cprint(forge.RED,0,'removing ' + dir)
				shutil.rmtree(dir)

		forge.cprint(forge.RED,0,'removing .pyc files')
		os.path.walk(forge.rootPath,scrub_tree,None)

		fullname = os.path.join(forge.rootPath,'test','log.txt')
		if os.path.exists(fullname):
			forge.cprint(forge.YELLOW,0,'removing: '+ fullname)
			os.remove(fullname)

		fullname = os.path.join(forge.srcPath,'fe','config.h')
		if os.path.exists(fullname):
			forge.cprint(forge.YELLOW,0,'removing: '+ fullname)
			os.remove(fullname)

		cwd = os.getcwd()
		os.chdir(forge.rootPath)
		os.system('svn status')
		forge.cprint(forge.WHITE,1,'x86_win32:')
		os.system('svn propget x86_win32 .')
		forge.cprint(forge.WHITE,1,'x86_linux:')
		os.system('svn propget x86_linux .')
		os.chdir(cwd)

scrubTarget = ScrubTarget("scrub")
forge.targetRegistry['scrub'] = scrubTarget

fe_help = """\
Free Electron specific help

FE Build Environment Variables:
FE_ROOT           - root directory of FE distribution
FE_LIB            - specify a particular directory for lib files
FE_OBJ            - specify a particular directory for object files
HASH_MAP_IS_MAP   - use std::map in place of hash_map
HASH_MAP_IS_MAP   - use std::map in place of hash_map
BOOST_INCLUDE     - location of the boost header directory
BOOST_LIB         - location of the boost header directory
BOOST_STRUCTURE   - (install or bin-stage)
BOOST_VERSION     - boost version (such as 1_31)
FE_CC             - compiler
FE_NO_UNIT        - disable automatic unit tests
FE_USE_SSE        - turn on SSE optimizations, if available
FE_USE_PCH        - turn on pre-compiled headers, if available
FE_E_BACKTRACE    - used at runtime, try to backtrace certain failures
FE_E_ATTACH       - used at runtime, try to attach to debugger on segfault
FE_E_ASSERT       - causes an assert in place of an exception
"""

forge.help.append(fe_help)


def dummy_execute(command,name):
	print command

class PretendTarget(forge.Target):
	def __init__(self,str):
		forge.Target.__init__(self,str)
		self.help = """\
Print out commands to be executed, but do not actually execute them.
"""
	def Build(self, reference_time):
		forge.execute = dummy_execute

pretendTarget = PretendTarget("pretend")
forge.targetRegistry['pretend'] = pretendTarget

def dot_build(self, reference_time):
	if self.buildNum >= forge.buildNum:
		return self.LastBuild()
	self.buildNum = forge.buildNum

	for dep in self.deps:
		(s_h,s_t) = os.path.split(self.name)
		(d_h,d_t) = os.path.split(dep.name)
		forge.dotfile.write('"%s" -> "%s"\n' % (s_t,d_t))
		if forge.recursionDepth < 0 or forge.recursionLevel < forge.recursionDepth:
			forge.recursionLevel += 1
			dep.Build(reference_time)

class DotFile:
	def __init__(self):
		self.file = file('out.dot', "w")
		self.file.write('digraph {\n')
	def __del__(self):
		self.file.write('}\n')
		self.file.close()
		import os
		os.system("dot -Tps out.dot -o out.ps")
	def write(self, s):
		self.file.write(s)

class DotTarget(forge.Target):
	def __init__(self,str):
		forge.Target.__init__(self,str)
		self.help = """\
Generate dep.dot, a graph of build dependencies
"""
	def Build(self, reference_time):
		forge.dotfile = DotFile()
		forge.Target.Build = dot_build

dotTarget = DotTarget("dot")
forge.targetRegistry['dot'] = dotTarget

def copy_headers(arg,dirname,names):
	(base_dir, include_dir) = arg
	re_compile = re.compile(r'.*\.(h|pmh)$')
	rel_path = forge.RelativePath(base_dir,dirname)
	if not rel_path:
		return
	target_dir = os.path.join(include_dir, rel_path)
	if not os.path.exists(target_dir):
		try:
			os.makedirs(target_dir)
		except OSError:
			print "failed creating " + target_dir
	for filename in names:
		if re_compile.match(filename):
			src = os.path.join(dirname, filename)
			dst = os.path.join(target_dir, filename)
			print("%s -> %s" % (src,dst))
			shutil.copyfile(src,dst)

class IncludeTarget(forge.Target):
	def __init__(self,str):
		forge.Target.__init__(self,str)
		self.help = """\
Extract header files to include directory
"""
	def Build(self, reference_time):
		include_dir = os.path.join(forge.rootPath, 'include')
		for src in ['src', 'ext']:
			src_path = os.path.join(forge.rootPath, src)
			os.path.walk(src_path,copy_headers,(src_path, include_dir))

includeTarget = IncludeTarget("include")
forge.targetRegistry['include'] = includeTarget


############################################################################
# xfig Image conversion support
def search_file(filename):
	found = 0
	os_path = os.getenv('PATH')
	if not os_path:
		os_path = os.defpath
	paths = string.split(os_path, os.pathsep)
	for path in paths:
		if os.path.exists(os.path.join(path, filename)):
			found = 1
			break
	if found:
		return os.path.abspath(os.path.join(path, filename))
	else:
		return None

def fig2jpg(target):
	forge.cprint(forge.WHITE,0,'jpg: fig2dev %s' % forge.RelativePath(forge.rootPath,target.name))
	prog = search_file('fig2dev')
	if prog:
		forge.confirm_path(target)
		command = "%s -L jpeg -S 4 -q 100 %s > %s" % (prog, target.deps[0].name,target.name)
		os.system(command)
	else:
		forge.cprint(forge.RED,0,'fig2dev not found')

def fig2eps(target):
	forge.cprint(forge.WHITE,0,'eps: fig2dev %s' % forge.RelativePath(forge.rootPath,target.name))
	prog = search_file('fig2dev')
	if prog:
		forge.confirm_path(target)
		command = "%s -L eps %s > %s" % (prog, target.deps[0].name,target.name)
		os.system(command)
	else:
		forge.cprint(forge.RED,0,'fig2dev not found')

def Module_Image(self, name):
	src_name = os.path.join(self.modPath,name + ".fig")
	jpg_name = os.path.join(forge.rootPath,'doc','image',name + ".jpg")
	eps_name = os.path.join(forge.rootPath,'doc','image',name + ".eps")
	src = forge.FileTarget(src_name)
	jpg = forge.FileTarget(jpg_name)
	eps = forge.FileTarget(eps_name)
	jpg.AddDep(src)
	eps.AddDep(src)
	self.AddDep(jpg)
	self.AddDep(eps)
	dox = forge.targetRegistry['documentation']
	dox.AddDep(jpg)
	dox.AddDep(eps)
	return (jpg, eps)

forge.defaultContext.AppendRule(r"(.*)\.(jpg)", fig2jpg)
forge.defaultContext.AppendRule(r"(.*)\.(eps)", fig2eps)

forge.Module.Image = Module_Image

##############################################################################
# prettify stl output
def stl_gcc_decrypt(raw):
	command = 'perl ' + os.path.join(forge.rootPath,'bin','gSTLFilt.pl')
	(p_stdin, p_stdout) = os.popen2(command)
	p_stdin.write(raw)
	p_stdin.close()
	cooked = p_stdout.read()
	n = p_stdout.close()
	return string.rstrip(cooked)

def colorize_filenames(raw):
	re_c = re.compile(r'^(|.*\s)(\S+\.(cc|h|c|pmh)):(\d*)(.*)$')
	lines = string.split(raw,'\n')
	cooked = ''
	for l in lines:
		m = re_c.match(l)
		if m:
			cooked += m.group(1)
			cooked += forge.ansi(forge.BOLD)
			cooked += forge.ansi(forge.BLUE)
			cooked += m.group(2)
			cooked += forge.ansi(forge.NORMAL) + ':' + forge.ansi(forge.WHITE)
			cooked += m.group(4)
			cooked += forge.ansi(forge.NORMAL)
			cooked += m.group(5)
			cooked += '\n'
		else:
			cooked += l + '\n'
	return string.rstrip(cooked)

def colorize_errors(raw):
	error = forge.ansi(forge.BOLD) + forge.ansi(forge.RED) + 'error' + forge.ansi(forge.NORMAL)
	cooked = string.replace(raw, 'error', error)
	return cooked

def colorize_warnings(raw):
	warning = forge.ansi(forge.BOLD) + forge.ansi(forge.YELLOW) + 'warning' + forge.ansi(forge.NORMAL)
	cooked = string.replace(raw, 'warning', warning)
	return cooked

if forge.fe_os == "FE_LINUX":
	# gSTLFilt.pl breaks on gcc 3.4
	if forge.cxx == "g++":
		forge.outputContext.AppendRule(r"(.*)\.(obj|o)", stl_gcc_decrypt)
	forge.outputContext.AppendRule(r"(.*)\.(obj|o)", colorize_filenames)
	forge.outputContext.AppendRule(r"(.*)\.(obj|o)", colorize_errors)
	forge.outputContext.AppendRule(r"(.*)\.(obj|o)", colorize_warnings)

##############################################################################

def bld_exists(rel_path):
	return os.path.exists(os.path.join(forge.rootPath, rel_path, 'bld.py'))

forge.tests = []

module_list	 = [ "src", "ext", "lab", "doc" ]

for i in module_list:
	if bld_exists(i):
		forge.setup(i)

if bld_exists('test'):
	test_module = forge.setup("test")
	if not os.environ.has_key("FE_NO_UNIT"):
		forge.dep(test_module, "unit")

if bld_exists('app'):
	lab_module =  forge.setup("app")

